# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""HTTP server for COS Alerter."""

import hashlib
import logging

import timeago
from flask import Flask, render_template, request
from prometheus_flask_exporter import PrometheusMetrics

from .alerter import AlerterState, config, now_datetime

app = Flask(__name__)
metrics = PrometheusMetrics(app)
logger = logging.getLogger(__name__)


@app.route("/", methods=["GET"])
def dashboard():
    """Endpoint for the COS Alerter dashboard."""
    clients = []
    now = now_datetime()
    for clientid in AlerterState.clients():
        with AlerterState(clientid) as state:
            last_alert = state.last_alert_datetime()
            alert_time = timeago.format(last_alert, now) if last_alert is not None else "never"
            status = "up" if not state.is_down() else "down"
            if last_alert is None:
                status = "unknown"
            client_name = config["watch"]["clients"][clientid].get("name", "")
            clients.append(
                {
                    "client_name": client_name,
                    "status": status,
                    "alert_time": alert_time,
                }
            )
    return render_template("dashboard.html", clients=clients)


@app.route("/alive", methods=["POST"])
def alive():
    """Endpoint for Alertmanager instances to send their heartbeat alerts."""
    # TODO Decide if we should validate the request.
    params = request.args
    clientid_list = params.getlist("clientid")  # params is a werkzeug.datastructures.MultiDict
    key_list = params.getlist("key")

    if len(clientid_list) < 1 or len(key_list) < 1:
        logger.warning("Request %s is missing clientid or key.", request.url)
        return 'Parameters "clientid" and "key" are required.', 400
    if len(clientid_list) > 1 or len(key_list) > 1:
        logger.warning("Request %s specified clientid or key more than once.", request.url)
        return 'Parameters "clientid" and "key" should be provided exactly once.', 400
    clientid = clientid_list[0]
    key = key_list[0]

    # Find the client with the specified clientid
    client_info = config["watch"]["clients"].get(clientid)
    if not client_info:
        logger.warning("Request %s specified an unknown clientid.", request.url)
        return 'Clientid {params["clientid"]} not found. ', 404

    # Hash the key and compare with the stored hashed key
    hashed_key = hashlib.sha512(key.encode()).hexdigest()
    if hashed_key != client_info.get("key", ""):
        logger.warning("Request %s provided an incorrect key.", request.url)
        return "Incorrect key for the specified clientid.", 401
    logger.info("Received alert from Alertmanager clientid: %s.", clientid)
    with AlerterState(clientid) as state:
        state.reset_alert_timeout()
    return "Success!"


@app.before_request
def log_request():
    """Log every HTTP request."""
    logger.info(
        "Request: %s %s",
        request.method,
        request.url,
    )



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""Daemon process for COS Alerter."""

import argparse
import logging
import signal
import sys
import threading
import time
from typing import List, Optional

import waitress

from .alerter import AlerterState, config, send_test_notification, up_time
from .logging import LEVELS, init_logging
from .server import app

logger = logging.getLogger("cos_alerter.daemon")


def sigint(_, __):  # pragma: no cover
    """Signal handler to exit cleanly on SIGINT."""
    logger.info("Received SIGINT.")
    logger.debug("Exiting.")
    sys.exit()


def sigterm(_, __):  # pragma: no cover
    """Signal handler for graceful shutdown on sigterm."""
    logger.info("Shutting down.")
    AlerterState.dump_and_pause()
    sys.exit()


def sigusr1(_, __):  # pragma: no cover
    """Signal handler for SIGUSR1 which sends a test notification."""
    logger.info("Received SIGUSR1.")
    send_thread = threading.Thread(target=send_test_notification)
    send_thread.start()


def parse_args(args: List[str]) -> argparse.Namespace:
    """Parse the command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log-level",
        choices=list(LEVELS),
        help='Logging level. Overrides config value "log_level"',
    )
    parser.add_argument(
        "--config",
        required=False,
        default="/etc/cos-alerter.yaml",
        help="Path to config file. Defaults to /etc/cos-alerter.yaml",
    )
    return parser.parse_args(args=args)


def client_loop(clientid):
    """Run the main loop for the specified client."""
    # Main loop
    state = AlerterState(clientid=clientid)
    while True:
        with state:
            logger.debug("Checking Alertmanager status.")
            if state.is_down():
                logger.debug("Alertmanager is down.")
                state.notify()
        time.sleep(1)


def main(run_for: Optional[int] = None, argv: List[str] = sys.argv):
    """Main method for COS Alerter.

    Args:
        run_for: This argument is for testing purposes. If set, only run for "run_for" seconds.
        argv: Can be used to override the cli args for testing purposes.
    """
    args = parse_args(argv[1:])

    config.set_path(args.config)
    config.reload()
    init_logging(args)
    AlerterState.initialize()

    # Observe signal handlers
    try:  # pragma: no cover
        signal.signal(signal.SIGINT, sigint)
        signal.signal(signal.SIGTERM, sigterm)
        signal.signal(signal.SIGUSR1, sigusr1)
        logger.debug("Signal handlers set.")
    except ValueError as e:
        # If we are not in the main thread, we can not start the signal handlers.
        # This is okay.
        if not str(e) == "signal only works in main thread of the main interpreter":
            raise  # pragma: no cover

    # Start the web server.
    # Starting in a thread rather than a new process allows waitress to inherit the log level
    # from the daemon. It also facilitates communication over memory rather than files.
    # clear_untrusted_proxy_headers is set to suppress a DeprecationWarning.
    server_thread = threading.Thread(
        target=waitress.serve,
        args=(app,),
        kwargs={
            "clear_untrusted_proxy_headers": True,
            "listen": config["web_listen_addr"],
        },
    )
    server_thread.daemon = True  # Makes this thread exit when the main thread exits.
    logger.info("Starting the web server thread.")
    server_thread.start()

    for clientid in config["watch"]["clients"]:
        client_thread = threading.Thread(target=client_loop, args=(clientid,))
        client_thread.daemon = True  # Makes this thread exit when the main thread exits.
        logger.info("Starting worker thread for client: %s", clientid)
        client_thread.start()

    while True:
        if run_for is not None and up_time() >= run_for:
            return
        time.sleep(1)


if __name__ == "__main__":
    main()  # pragma: no cover



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

import os
import textwrap
import threading
import unittest.mock

import apprise
import freezegun
import yaml
from helpers import DESTINATIONS

from cos_alerter.alerter import AlerterState, config, send_test_notification, up_time


def assert_notifications(notify_mock, add_mock, title, body):
    add_mock.assert_has_calls([unittest.mock.call(x) for x in DESTINATIONS])
    notify_mock.assert_called_with(title=title, body=body)


def test_config_gets_item(fake_fs):
    assert config["watch"]["down_interval"] == 300


def test_config_default_empty_file(fake_fs):
    with open("/etc/cos-alerter.yaml", "w") as f:
        f.write("")
    config.reload()
    assert config["watch"]["down_interval"] == 300


def test_file_not_found_error(fake_fs):
    with unittest.mock.patch("cos_alerter.alerter.logger") as mock_logger:
        os.unlink("/etc/cos-alerter.yaml")
        try:
            config.reload()
        except SystemExit as exc:
            assert exc.code == 1
            mock_logger.critical.assert_called_once_with("Config file not found. Exiting...")
        else:
            assert False


def test_duplicate_key_error(fake_fs):
    duplicate_config = """
    watch:
      down_interval: "5m"
      wait_for_first_connection: true
      clients:
        clientid1:
          key: "clientkey1"
          name: "Instance Name 1"
        clientid1:
          key: "clientkey1"
          name: "Instance Name 1"
    """
    with open("/etc/cos-alerter.yaml", "w") as f:
        f.write(duplicate_config)

    try:
        config.reload()
    except SystemExit as exc:
        assert exc.code == 1
    else:
        # If no exception is raised, fail the test
        assert False


def test_invalid_hashes(fake_fs):
    duplicate_config = """
    watch:
      down_interval: "5m"
      wait_for_first_connection: true
      clients:
        invalidhashclient:
          key: "E0E06B8DB6ED8DD4E1FFE98376E606BDF4FE4ABB4AF65BFE8B18FBFA6564D8B3"
          name: "Instance Name 1"
    """
    with open("/etc/cos-alerter.yaml", "w") as f:
        f.write(duplicate_config)

    try:
        config.reload()
    except SystemExit as exc:
        assert exc.code == 1
    else:
        # If no exception is raised, fail the test
        assert False


def test_config_default_partial_file(fake_fs):
    conf = yaml.dump({"log_level": "info"})
    with open("/etc/cos-alerter.yaml", "w") as f:
        f.write(conf)
    config.reload()
    assert config["watch"]["down_interval"] == 300


def test_config_default_override(fake_fs):
    conf = yaml.dump({"watch": {"down_interval": "1m"}})
    with open("/etc/cos-alerter.yaml", "w") as f:
        f.write(conf)
    config.reload()
    assert config["watch"]["down_interval"] == 60


@freezegun.freeze_time("2023-01-01")
@unittest.mock.patch("time.monotonic")
def test_initialize(monotonic_mock, fake_fs):
    monotonic_mock.return_value = 1000
    AlerterState.initialize()
    state = AlerterState(clientid="clientid1")
    with state:
        assert state.start_date == 1672531200.0
        assert state.start_time == 1000
        assert state.data["alert_time"] == 1000
        assert state.data["notify_time"] is None


@freezegun.freeze_time("2023-01-01")
@unittest.mock.patch("time.monotonic")
def test_up_time(monotonic_mock, fake_fs):
    monotonic_mock.return_value = 1000
    AlerterState.initialize()
    monotonic_mock.return_value = 2000
    assert up_time() == 1000


@freezegun.freeze_time("2023-01-01")
@unittest.mock.patch("time.monotonic")
def test_is_down_from_initialize(monotonic_mock, fake_fs):
    monotonic_mock.return_value = 1000
    AlerterState.initialize()
    state = AlerterState(clientid="clientid1")
    with state:
        monotonic_mock.return_value = 1180  # Three minutes have passed
        assert state.is_down() is False
        monotonic_mock.return_value = 1330  # Five and a half minutes have passed
        assert state.is_down() is True


@freezegun.freeze_time("2023-01-01")
@unittest.mock.patch("time.monotonic")
def test_is_down_with_reset_alert_timeout(monotonic_mock, fake_fs):
    monotonic_mock.return_value = 1000
    AlerterState.initialize()
    state = AlerterState(clientid="clientid1")
    with state:
        monotonic_mock.return_value = 2000
        state.reset_alert_timeout()
        monotonic_mock.return_value = 2180  # Three minutes have passed
        assert state.is_down() is False
        monotonic_mock.return_value = 2330  # Five and a half minutes have passed
        assert state.is_down() is True


@freezegun.freeze_time("2023-01-01")
@unittest.mock.patch("time.monotonic")
def test_is_down_with_wait_for_first_connection(monotonic_mock, fake_fs):
    with open("/etc/cos-alerter.yaml") as f:
        conf = yaml.safe_load(f)
    conf["watch"]["wait_for_first_connection"] = True
    with open("/etc/cos-alerter.yaml", "w") as f:
        yaml.dump(conf, f)
    config.reload()
    monotonic_mock.return_value = 1000
    AlerterState.initialize()
    state = AlerterState(clientid="clientid1")
    with state:
        monotonic_mock.return_value = 1500
        assert state.is_down() is False  # 6 minutes have passes but we have not started counting.
        state.reset_alert_timeout()
        monotonic_mock.return_value = 1680
        assert state.is_down() is False  # Three more minutes.
        monotonic_mock.return_value = 1830
        assert state.is_down() is True  # 5.5 minutes since reset_alert_timeout() was called.


@freezegun.freeze_time("2023-01-01")
@unittest.mock.patch("time.monotonic")
def test_is_down_from_graceful_shutdown(monotonic_mock, fake_fs):
    with open("/etc/cos-alerter.yaml") as f:
        conf = yaml.safe_load(f)
    conf["watch"]["wait_for_first_connection"] = True
    with open("/etc/cos-alerter.yaml", "w") as f:
        yaml.dump(conf, f)
    config.reload()
    fake_fs.create_file(config["clients_file"])
    with config["clients_file"].open("w") as f:
        f.write('{"clientid1": {"alert_time": 500, "notify_time": null}}')
    monotonic_mock.return_value = 1000
    print("Hello Test")
    AlerterState.initialize()
    state = AlerterState(clientid="clientid1")
    with state:
        print(list(AlerterState.clients()))
        print(state.data)
        assert state.is_down() is False
        monotonic_mock.return_value = 2330
        assert state.is_down() is True


@freezegun.freeze_time("2023-01-01")
@unittest.mock.patch("time.monotonic")
def test_is_down(monotonic_mock, fake_fs):
    monotonic_mock.return_value = 1000
    AlerterState.initialize()
    state = AlerterState(clientid="clientid1")
    with state:
        monotonic_mock.return_value = 2000
        state.reset_alert_timeout()
        monotonic_mock.return_value = 2180  # Three minutes have passed
        assert state.is_down() is False
        monotonic_mock.return_value = 2330  # Five and a half minutes have passed
        assert state.is_down() is True


@freezegun.freeze_time("2023-01-01")
@unittest.mock.patch("time.monotonic")
def test_recently_notified(monotonic_mock, fake_fs):
    monotonic_mock.return_value = 1000
    AlerterState.initialize()
    state = AlerterState(clientid="clientid1")
    with state:
        state._set_notify_time()
        monotonic_mock.return_value = 2800  # 30 minutes have passed
        assert state._recently_notified() is True
        monotonic_mock.return_value = 5200  # 70 minutes have passed
        assert state._recently_notified() is False


@freezegun.freeze_time("2023-01-01")
@unittest.mock.patch("time.monotonic")
@unittest.mock.patch.object(apprise.Apprise, "add")
@unittest.mock.patch.object(apprise.Apprise, "notify")
def test_notify(notify_mock, add_mock, monotonic_mock, fake_fs):
    monotonic_mock.return_value = 1000
    AlerterState.initialize()
    state = AlerterState(clientid="clientid1")

    with state:
        state.notify()
    for thread in threading.enumerate():
        if thread != threading.current_thread():
            thread.join()
    assert_notifications(
        notify_mock,
        add_mock,
        title="**Alertmanager is Down!**",
        body=textwrap.dedent(
            """
            Your Alertmanager instance: clientid1 seems to be down!
            It has not alerted COS-Alerter ever.
            """
        ),
    )

    # Make sure if we try again, nothing is sent
    notify_mock.reset_mock()

    with state:
        state.notify()
    for thread in threading.enumerate():
        if thread != threading.current_thread():
            thread.join()
    notify_mock.assert_not_called()


@unittest.mock.patch.object(apprise.Apprise, "add")
@unittest.mock.patch.object(apprise.Apprise, "notify")
def test_send_test_notification(notify_mock, add_mock, fake_fs):
    send_test_notification()
    assert_notifications(
        notify_mock,
        add_mock,
        title="COS-Alerter test email.",
        body="This is a test email automatically generated by COS-alerter.",
    )
